@@ -1,4 +1,7 @@
-use crate::{port_is_unique, Error, Read, Seek, SeekFrom};
+use crate::{
+    io::{Read, Seek, SeekFrom, Write},
+    port_is_unique, Error,
+};
 use core::cmp::min;
 use w5500_ll::{
     net::SocketAddrV4, Protocol, Registers, Sn, SocketCommand, SocketMode, SocketStatus,
@@ -20,7 +23,7 @@ use w5500_ll::{
 ///     net::{Ipv4Addr, SocketAddrV4},
 ///     Tcp,
 ///     TcpReader,
-///     Read,
+///     io::Read,
 /// };
 ///
 /// const MQTT_SOCKET: Sn = Sn::Sn0;
@@ -56,16 +59,16 @@ use w5500_ll::{
 /// ```
 #[derive(Debug)]
 #[cfg_attr(feature = "defmt", derive(defmt::Format))]
-pub struct TcpReader<'a, W: Registers> {
-    pub(crate) w5500: &'a mut W,
+pub struct TcpReader<'w, W5500: Registers> {
+    pub(crate) w5500: &'w mut W5500,
     pub(crate) sn: Sn,
     pub(crate) head_ptr: u16,
     pub(crate) tail_ptr: u16,
     pub(crate) ptr: u16,
 }
 
-impl<'a, W: Registers> Seek<W::Error> for TcpReader<'a, W> {
-    fn seek(&mut self, pos: SeekFrom) -> Result<(), Error<W::Error>> {
+impl<'w, W5500: Registers> Seek<W5500::Error> for TcpReader<'w, W5500> {
+    fn seek(&mut self, pos: SeekFrom) -> Result<(), Error<W5500::Error>> {
         self.ptr = pos.new_ptr(self.ptr, self.head_ptr, self.tail_ptr)?;
         Ok(())
     }
@@ -87,8 +90,8 @@ impl<'a, W: Registers> Seek<W::Error> for TcpReader<'a, W> {
     }
 }
 
-impl<'a, W: Registers> Read<'a, W> for TcpReader<'a, W> {
-    fn read(&mut self, buf: &mut [u8]) -> Result<u16, W::Error> {
+impl<'a, W5500: Registers> Read<W5500::Error> for TcpReader<'a, W5500> {
+    fn read(&mut self, buf: &mut [u8]) -> Result<u16, W5500::Error> {
         let read_size: u16 = min(self.remain(), buf.len().try_into().unwrap_or(u16::MAX));
         if read_size != 0 {
             self.w5500
@@ -101,7 +104,7 @@ impl<'a, W: Registers> Read<'a, W> for TcpReader<'a, W> {
         }
     }
 
-    fn read_exact(&mut self, buf: &mut [u8]) -> Result<(), Error<W::Error>> {
+    fn read_exact(&mut self, buf: &mut [u8]) -> Result<(), Error<W5500::Error>> {
         let buf_len: u16 = buf.len().try_into().unwrap_or(u16::MAX);
         let read_size: u16 = min(self.remain(), buf_len);
         if read_size != buf_len {
@@ -113,13 +116,112 @@ impl<'a, W: Registers> Read<'a, W> for TcpReader<'a, W> {
         }
     }
 
-    fn done(self) -> Result<(), W::Error> {
+    fn done(self) -> Result<(), W5500::Error> {
         self.w5500.set_sn_rx_rd(self.sn, self.ptr)?;
         self.w5500.set_sn_cr(self.sn, SocketCommand::Recv)?;
         Ok(())
     }
 }
 
+/// Streaming writer for a UDP socket buffer.
+///
+/// This implements the [`Seek`] traits.
+///
+/// Created with [`Udp::udp_writer`](crate::Udp::udp_writer).
+///
+/// # Example
+///
+/// ```no_run
+/// # use embedded_hal_mock as h;
+/// # let mut w5500 = w5500_ll::blocking::vdm::W5500::new(h::spi::Mock::new(&[]), h::pin::Mock::new(&[]));
+/// use w5500_hl::{
+///     ll::{Registers, Sn, SocketInterrupt},
+///     net::{Ipv4Addr, SocketAddrV4},
+///     Tcp, TcpWriter, io::Write,
+/// };
+///
+/// const MQTT_SOCKET: Sn = Sn::Sn0;
+/// const MQTT_SOURCE_PORT: u16 = 33650;
+/// const MQTT_SERVER: SocketAddrV4 = SocketAddrV4::new(Ipv4Addr::new(192, 168, 2, 10), 1883);
+///
+/// w5500.tcp_connect(MQTT_SOCKET, MQTT_SOURCE_PORT, &MQTT_SERVER)?;
+///
+/// // ... wait for a CON interrupt
+///
+/// const CONNECT: [u8; 14] = [
+///     0x10, 0x0C, 0x00, 0x04, b'M', b'Q', b'T', b'T', 0x04, 0x02, 0x0E, 0x10, 0x00, 0x00,
+/// ];
+/// let mut writer: TcpWriter<_> = w5500.tcp_writer(MQTT_SOCKET)?;
+/// writer.write_all(&CONNECT)?;
+/// writer.send()?;
+/// # Ok::<(), w5500_hl::Error<_>>(())
+/// ```
+#[derive(Debug, PartialEq, Eq)]
+#[cfg_attr(feature = "defmt", derive(defmt::Format))]
+pub struct TcpWriter<'w, W5500: Registers> {
+    pub(crate) w5500: &'w mut W5500,
+    pub(crate) sn: Sn,
+    pub(crate) head_ptr: u16,
+    pub(crate) tail_ptr: u16,
+    pub(crate) ptr: u16,
+}
+
+impl<'w, W5500: Registers> Seek<W5500::Error> for TcpWriter<'w, W5500> {
+    fn seek(&mut self, pos: SeekFrom) -> Result<(), Error<W5500::Error>> {
+        self.ptr = pos.new_ptr(self.ptr, self.head_ptr, self.tail_ptr)?;
+        Ok(())
+    }
+
+    fn rewind(&mut self) {
+        self.ptr = self.head_ptr
+    }
+
+    fn stream_len(&self) -> u16 {
+        self.tail_ptr.wrapping_sub(self.head_ptr)
+    }
+
+    fn stream_position(&self) -> u16 {
+        self.ptr.wrapping_sub(self.head_ptr)
+    }
+
+    fn remain(&self) -> u16 {
+        self.tail_ptr.wrapping_sub(self.ptr)
+    }
+}
+
+impl<'w, W5500: Registers> Write<W5500::Error> for TcpWriter<'w, W5500> {
+    fn write(&mut self, buf: &[u8]) -> Result<u16, W5500::Error> {
+        let write_size: u16 = min(self.remain(), buf.len().try_into().unwrap_or(u16::MAX));
+        if write_size != 0 {
+            self.w5500
+                .set_sn_tx_buf(self.sn, self.ptr, &buf[..usize::from(write_size)])?;
+            self.ptr = self.ptr.wrapping_add(write_size);
+
+            Ok(write_size)
+        } else {
+            Ok(0)
+        }
+    }
+
+    fn write_all(&mut self, buf: &[u8]) -> Result<(), Error<W5500::Error>> {
+        let buf_len: u16 = buf.len().try_into().unwrap_or(u16::MAX);
+        let write_size: u16 = min(self.remain(), buf_len);
+        if write_size != buf_len {
+            Err(Error::OutOfMemory)
+        } else {
+            self.w5500.set_sn_tx_buf(self.sn, self.ptr, buf)?;
+            self.ptr = self.ptr.wrapping_add(write_size);
+            Ok(())
+        }
+    }
+
+    fn send(self) -> Result<(), W5500::Error> {
+        self.w5500.set_sn_tx_wr(self.sn, self.ptr)?;
+        self.w5500.set_sn_cr(self.sn, SocketCommand::Send)?;
+        Ok(())
+    }
+}
+
 /// A W5500 TCP trait.
 pub trait Tcp: Registers {
     /// Starts the 3-way TCP handshake with the remote host.
@@ -199,11 +301,8 @@ pub trait Tcp: Registers {
         // This will not hang, the socket status will always change to closed
         // after a close command.
         // (unless you do somthing silly like holding the W5500 in reset)
-        loop {
-            if self.sn_sr(sn)? == Ok(SocketStatus::Closed) {
-                break;
-            }
-        }
+        while self.sn_sr(sn)? != Ok(SocketStatus::Closed) {}
+
         const MODE: SocketMode = SocketMode::DEFAULT.set_protocol(Protocol::Tcp);
         self.set_sn_mr(sn, MODE)?;
         self.set_sn_port(sn, port)?;
@@ -212,11 +311,8 @@ pub trait Tcp: Registers {
         // This will not hang, the socket status will always change to Init
         // after a open command with SN_MR set to TCP.
         // (unless you do somthing silly like holding the W5500 in reset)
-        loop {
-            if self.sn_sr(sn)? == Ok(SocketStatus::Init) {
-                break;
-            }
-        }
+        while self.sn_sr(sn)? != Ok(SocketStatus::Init) {}
+
         self.set_sn_cr(sn, SocketCommand::Connect)
     }
 
@@ -293,11 +389,7 @@ pub trait Tcp: Registers {
         // This will not hang, the socket status will always change to closed
         // after a close command.
         // (unless you do somthing silly like holding the W5500 in reset)
-        loop {
-            if self.sn_sr(sn)? == Ok(SocketStatus::Closed) {
-                break;
-            }
-        }
+        while self.sn_sr(sn)? != Ok(SocketStatus::Closed) {}
         const MODE: SocketMode = SocketMode::DEFAULT.set_protocol(Protocol::Tcp);
         self.set_sn_mr(sn, MODE)?;
         self.set_sn_port(sn, port)?;
@@ -305,11 +397,7 @@ pub trait Tcp: Registers {
         // This will not hang, the socket status will always change to Init
         // after a open command with SN_MR set to TCP.
         // (unless you do somthing silly like holding the W5500 in reset)
-        loop {
-            if self.sn_sr(sn)? == Ok(SocketStatus::Init) {
-                break;
-            }
-        }
+        while self.sn_sr(sn)? != Ok(SocketStatus::Init) {}
         self.set_sn_cr(sn, SocketCommand::Listen)
     }
 
@@ -521,6 +609,30 @@ pub trait Tcp: Registers {
             ptr: sn_rx_rd,
         })
     }
+
+    /// Create a TCP writer.
+    ///
+    /// This returns a [`TcpWriter`] structure, which contains functions to
+    /// stream data to the W5500 socket buffers incrementally.
+    ///
+    /// # Example
+    ///
+    /// See [`TcpWriter`].
+    fn tcp_writer(&mut self, sn: Sn) -> Result<TcpWriter<Self>, Self::Error>
+    where
+        Self: Sized,
+    {
+        let sn_tx_fsr: u16 = self.sn_tx_fsr(sn)?;
+        let sn_tx_wr: u16 = self.sn_tx_wr(sn)?;
+
+        Ok(TcpWriter {
+            w5500: self,
+            sn,
+            head_ptr: sn_tx_wr,
+            tail_ptr: sn_tx_wr.wrapping_add(sn_tx_fsr),
+            ptr: sn_tx_wr,
+        })
+    }
 }
 
 /// Implement the TCP trait for any structure that implements [`w5500_ll::Registers`].
